/*
 * Copyright Eiichi Yoshida, AIST
 */

#ifndef KINERH1_H
#define KINERH1_H

#include <vector>
#include <string>

#include "RH1Controller/tvmet3d.h"

#include "RH1Controller/RH1Defs.h"


using namespace OpenHRP;

class CkineRH1 {

 public:

  /**
     \brief constructor
  */
  CkineRH1(const std::string &i_paramFile = "RH1param.dat", 
	 unsigned int i_jointNum = DOF_RH1);

  /**
     \brief destructor
  */
  ~CkineRH1();

  /**
     \brief compute the IK.
     \brief input vector: x, y, z and r, p, y
     \return o_angleVec: all the solutions found within the limit.
     \return true if succeeded
  */

  bool computeIK(const std::vector<double> &i_goal, 
		 std::vector<std::vector<double> > &o_angleVec);

  /**
     \brief compute the IK.
     \brief input vector: x, y, z and r, p, y
     \return one solution closest to the initConf
     \return true if succeeded
  */
  bool computeIK(const std::vector<double> &i_goal, std::vector<double> &o_angle);

  /**
     \brief compute the FK with the angles from the base position.
     \brief input vector: joint angles
     \return true if succeeded
  */
  bool computeFK(const std::vector<double> &i_jointAngles);

  /**
     \brief get the posiiton and orientation of the joint 
     \param i_joint: joint angle
     \return pos: position of the joint
     \return rot: orientation of the joint
  */
  bool getJointPosFK(unsigned int i_joint, vector3 &pos, matrix33 &rot); 

  /**
     \brief load all the initial values of paramters from file 
     \brief (joint limits, link lengths)
     \param i_paramFile parameter file name
     \return true if all the papameters are set.
  */
  bool loadParam(const std::string &i_paramFile);

  /**
     \brief set the link length
     \param i_L: vector of link length
  */
  void L(const std::vector<double> &i_L);

  /**
     \brief set the init configuration
     \param i_initConf: vector of init config
  */
  void initConf(const std::vector<double> &i_initConf);

  /**
     \brief set the init configuration
     \param i_jointAngles: vector of init config
  */
  std::vector<double> jointAngles()  const { return m_jointAngles; };

  /**
     \brief set the lower joint limit 
     \param i_lLimit: vector of lower limit
  */
  void lLimit(const std::vector<double> &i_lLimit);

  /**
     \brief set the lower joint limit 
     \param i_lLimit: vector of lower limit
  */
  void uLimit(const std::vector<double> &i_uLimit);

  /**
     \brief set the lower joint limit 
     \param i_lLimit: vector of lower limit
  */
  void tcp(const vector3 &i_tcp) { m_tcp = i_tcp; } ;

  /**
     \brief get the angle between 3D vectors v1 and v2
     \brief the angle is positive from v1 to v2
     \param v1, v2: 3D vector 
     \return the angle
  */
  static double getVectorAngle(const std::vector<double> &i_v1, const std::vector<double> &i_v2);

  /**
     \brief check the joint limit
     \param i_jnum: joint nubmer
     \param i_val: joint value
  */
  bool checkLimit(unsigned int i_jnum, double i_val) const;

  /**
     \brief limit the angle between -pi and pi
     \return normalized angle.
  */

  static double limitAngle(double i_angle);

  /**
     \brief get parameter from file
     \param i_fileName: file name to read the param
     \param i_paramname: parameter name
     \param param: parameter read
  */
  template <class T> 
    bool getParamData(const std::string &i_fileName, unsigned int i_size,
		      const std::string &i_paramName, std::vector<T> &o_param);

  /**
     \brief sort the vector by the distance to init config
     \param io_angleVector: the vectors of angles to be sorted 
  */
  bool sortConfigFromInit(std::vector<std::vector<double> > &io_angleVector);


  /**
     \brief joint number
  */
  unsigned int m_jointNum;

  /**
     \brief link length    L0 (base to L1) ... L6 (J6 to End-effector)
  */
  std::vector<double> m_L;

  /**
     \brief initial configuration
  */
  std::vector<double> m_initConf;

  /**
     \brief joint angles for FK
  */
  std::vector<double> m_jointAngles;

  /**
     \brief file for parameters
  */
  std::string m_paramFile;

  /**
     \brief lower and upper limit 
  */
  std::vector<double> m_lLimit, m_uLimit;

  /**
     \brief vector to J6 to TCP
  */
  vector3 m_tcp;

  /**
     \brief limit set or not
  */
  bool m_limitFlag;

  /**
     \brief
  */
  bool m_fkFlag;

  /**
     \brief vector of joint axis directions
     0: X axis, 1: Y axis, 2: Z axis
  */
  std::vector<unsigned int> m_axisDir;

  /**
     \brief vector of joint positions
  */
  std::vector<vector3> m_jointPos;

  /**
     \brief vector of joint positions
  */
  std::vector<matrix33> m_jointOri;

};


#endif
