/*
 * Copyright Eiichi Yoshida, AIST
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <functional>
#include <algorithm>
#include <cassert>

#include "RH1Controller/tvmet3d.h"
#include "RH1Controller/kineRH1.h"

using namespace OpenHRP;

double L_RH1[DOF_RH1] = {0.1215, 0.1215, 0.25, 0.103, 0.09, 0.058};

// double TCP_RH1[3] = {0.00825, 0, 0.0475};
double TCP_RH1[3] = {0.00825, 0, 0.1};

// attention about the matrix operations with tvmet
// NG
// v = m * v
// 
// GOOD
// v = m * vector(v);
// v = vector(m * v);

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cerr << "CkineRH1:" << x << std::endl
#define ODEBUG1(x) std::cerr << "CkineRH1:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "CkineRH1:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

typedef std::pair<double, std::vector<double> > configPair;

struct compDist : 
  public std::binary_function<const configPair &, const configPair &, bool>
{
  bool operator()(const configPair &x, const configPair &y) const {
    return x.first < y.first;
  }
};



CkineRH1::CkineRH1(const std::string &i_paramFile, unsigned int i_jointNum) :
  m_jointNum(i_jointNum), m_initConf(std::vector<double>(DOF_RH1)),
  m_paramFile(i_paramFile), m_limitFlag(false),
  m_fkFlag(false)
{
  m_L.resize(m_jointNum);
  for(unsigned i=0; i<m_jointNum; i++)
    m_L[i] = L_RH1[i];

  m_tcp = OpenHRP::vector3(TCP_RH1[0], TCP_RH1[1], TCP_RH1[2]);

  // get the parameters
  if(!loadParam(i_paramFile)){
    ODEBUG1("constructor: loading parameter from "<<i_paramFile
	    <<" not succeeded. ");
  }
}

CkineRH1::~CkineRH1() 
{
}


bool CkineRH1::computeIK(const std::vector<double> &i_goal, 
		       std::vector<std::vector<double> > &o_angleVec)
{
  // vectors of each angle
  // theta[0] = vector -- solutions for theta1  (2 possibilities) 
  // theta[1] = vector -- solutions for theta2 ...
  // ...
  // theta[1],[2] are dependent
  std::vector<std::vector<double> > theta(m_jointNum);
  for(unsigned int i=0; i<theta.size(); i++)
    theta[i].clear();

  o_angleVec.clear();

  // get the target position
  matrix33 RGoal;
  RGoal = OpenHRP::rotFromRpy(i_goal[3], i_goal[4], i_goal[5]);

  // for the IK, correct the origin to center of J2
  // we should subtract L0 and L1 from z

  vector3 vGoal (i_goal[0], i_goal[1], i_goal[2] - (m_L[0]+m_L[1]));
  ODEBUG2("computeIK(): vGoal: "<<vGoal);
  ODEBUG2("computeIK(): RGoal: "<<RGoal);

  // The imput is supposed to be from the base of the arm


  // get the J6 position P6 and orientation 
  // P6 = goal - R*tcp
  
  vector3 tcp_abs;
  tcp_abs = RGoal * m_tcp;
  vector3 p6;
  p6 = vGoal - tcp_abs;

  ODEBUG2("computeIK(): p6: "<<p6);

  // get the position p5
  vector3 e6y, e5z;
  getVector3<matrix33>(e5z, RGoal, 0, ZCOL);
  getVector3<matrix33>(e6y, RGoal, 0, YCOL);

  vector3 p5;
  p5 = p6 - m_L[5] * e5z;

  ODEBUG2("computeIK(): e6y: "<<e6y<<", e5z: "<<e5z<<", p5: "<<p5);

  std::vector<double> theta1_tmp;
  theta1_tmp.clear();
  // get the angle theta 1 (two possible solutions, verified with joint limit)
  // get the angle between -PI and PI
  double theta11 = atan2(p5[1], p5[0]);
  ODEBUG2("computeIK(): theta11 "<<rad2deg(theta11));
  if(checkLimit(ID_J1, theta11)){
    theta1_tmp.push_back(theta11);
    ODEBUG2("computeIK(): limit OK "<<rad2deg(theta11));
  }
  double theta12 = limitAngle(theta11+PI);
  ODEBUG2("computeIK(): theta12 "<<rad2deg(theta12));
  if(checkLimit(ID_J1, theta12)){
    theta1_tmp.push_back(theta12);
    ODEBUG2("computeIK(): limit OK "<<rad2deg(theta12));
  }

  ODEBUG2("computeIK(): theta 1 has "<<theta1_tmp.size()<<" solutions. ");
  
  // see if there are solutions
  if(theta1_tmp.size() == 0){
    ODEBUG2("for theta1, no solution: return error");
    return false;
  }

  // get theta2 and theta3 (signs )
  // eq. (x1.8) - (x1.10) in Takano's book
  double p5len = norm2(p5);
  double p5len_sq = p5len * p5len;
  double l2_sq = m_L[2]*m_L[2];
  double l34 = m_L[3]+m_L[4];
  double l34_sq = l34*l34;

  ODEBUG2("computeIK(): p5len "<<p5len<<", l2 "<<m_L[2]<<", l3+l4 "<<l34
	  <<", l2+l34 "<<m_L[2]+l34 );
  if(p5len > m_L[2] + l34){
    ODEBUG1("computeIK(): cannot form triangle for theta 2 and 3");
    return false;
  }

  // alpha is different according to theta 1
  for(unsigned int i=0; i<theta1_tmp.size(); i++){

    double theta1 = theta1_tmp[i];
    ODEBUG2("computeIK(): "<<i<<"th theta1: "<<theta1);

    // check the sign of alpha
    // get the new x axis after theta 1 rotation
    vector3 e2x(cos(theta1), sin(theta1), 0);
    double sign_alpha = dot(e2x, p5) > 0 ? 1.0 : -1.0;
    ODEBUG2("e2x "<<e2x);
    ODEBUG2("dot (ex2, p5) "<<dot(e2x, p5)<<", sign alpha "<<sign_alpha);

    double alpha = atan2(sign_alpha*sqrt(p5[0]*p5[0]+p5[1]*p5[1]), p5[2]);
    double cbeta = (l2_sq + p5len_sq - l34_sq)/(2.0 * m_L[2] * p5len);
    double beta = acos(cbeta);
    double cgamma = (l34_sq + p5len_sq - l2_sq)/(2.0 * l34 * p5len);
    double gamma = acos(cgamma);

    ODEBUG2("computeIK(): cbeta "<<cbeta<<", cgamma "<<cgamma);
    ODEBUG2("computeIK(): for theta 2 and 3: alpha "<<rad2deg(alpha)
	  <<", beta "<<rad2deg(beta)<<", gamma "<<rad2deg(gamma));

    // solutions
    double theta21 = alpha-beta;
    double theta22 = alpha+beta;
    double theta31 = gamma+beta;
    double theta32 = -(gamma+beta);
    ODEBUG2("computeIK(): sol 1: (theta2, theta3): "<<rad2deg(theta21)<<", "
	  <<rad2deg(theta31));
    ODEBUG2("computeIK(): sol 2: (theta2, theta3): "<<rad2deg(theta22)<<", "
	    <<rad2deg(theta32));

    // check limit
    if(checkLimit(ID_J2, theta21) && checkLimit(ID_J3, theta31)){
      ODEBUG2("computeIK(): sol 1 valid. ");
      theta[0].push_back(theta1);
      theta[1].push_back(theta21);
      theta[2].push_back(theta31);
    }
    if(checkLimit(ID_J2, theta22) && checkLimit(ID_J3, theta32)){
      ODEBUG2("computeIK(): sol 2 valid. ");
      theta[0].push_back(theta1);
      theta[1].push_back(theta22);
      theta[2].push_back(theta32);
    }
    
  }
  if(theta[0].size() == 0){
    ODEBUG2("for theta2, no solution: return error");
    return false;
  }

  // get the position J3
  // there are maximum 4 (theta1, theta2, theta3) solutions
  for(unsigned int i=0; i<theta[0].size(); i++){

    double theta1 = theta[0][i];
    double theta2 = theta[1][i];
    double theta3 = theta[2][i];
    ODEBUG2("computeIK(): i "<<i<<": theta1 "<<rad2deg(theta1)<<", "
	    <<"theta2 "<<rad2deg(theta2)<<", "<<"theta3 "<<rad2deg(theta3));
    vector3 e3y(-sin(theta1), cos(theta1), 0);
    ODEBUG2("computeIK(): i "<<i<<", e3y "<<e3y);

    vector3 p3(m_L[2]*sin(theta2)*cos(theta1), m_L[2]*sin(theta2)*sin(theta1),
	       m_L[2]*cos(theta2));

    // vector p3 -> p5
    vector3 p35;
    p35 = p5 - p3;
    ODEBUG2("computeIK(): norm p35" <<norm2(p35)<<", l3+l4 "<<m_L[3]+m_L[4]);
    vector3 e3z;
    e3z = normalize(p35);
    ODEBUG2("computeIK(): p35 "<<p35<<", e3z "<<e3z);
    // 2 possibilites of e3y
    for(unsigned k=0; k<2; k++){
      ODEBUG2("computeIK(): k "<<k);
      vector3 v5y, e5y;
      v5y = (1.0-2*k)*cross(e3z, e5z); // k=0: 1, k=1: -1
      e5y = normalize(v5y);
      ODEBUG2("computeIK():  e3y "<<e3y);
	
      double c5 = dot(e3z, e5z);
      double s5 = dot(cross(e3z,e5z), e5y);
      double theta5 = atan2(s5, c5);
      ODEBUG2("theta 5: "<<rad2deg(theta5));

      // check limit
      if(!checkLimit(ID_J5, theta5)){
	ODEBUG2("computeIK(): theta 5 not valid, try next possibility");
	continue;
      }

      double c4 = dot(e3y, e5y);
      double s4 = dot(cross(e3y, e5y), e3z);
      double theta4 = atan2(s4, c4);
      ODEBUG2("theta 4: "<<rad2deg(theta4));

      // check limit
      if(!checkLimit(ID_J4, theta4)){
	ODEBUG2("computeIK(): theta 4 not valid, try next possibility");
	continue;
      }

      double c6 = dot(e5y, e6y);
      double s6 = dot(cross(e5y, e6y), e5z);
      double theta6 = atan2(s6, c6);
      ODEBUG2("theta 6: "<<rad2deg(theta6));

      // check limit
      if(!checkLimit(ID_J6, theta6)){
	ODEBUG2("computeIK(): theta 6 not valid, try next possibility");
	continue;
      }

      ODEBUG2("computeIK(): theta 4, 5, 6 valid. add as solution");
      theta[3].push_back(theta4);
      theta[4].push_back(theta5);
      theta[5].push_back(theta6);

      // compute theta 5
      // make the solution vector
      std::vector<double> sol(m_jointNum);
      sol[0] = theta1;
      sol[1] = theta2;
      sol[2] = theta3;
      sol[3] = theta4;
      sol[4] = theta5;
      sol[5] = theta6;

      ODEBUG2("computeIK(): "<<o_angleVec.size()<<" th solution: ");
#if DEBUG==2
      for(unsigned int l=0; l<sol.size(); l++)
	std::cerr<<" "<<rad2deg(sol[l]);
      std::cerr<<std::endl;
#endif	
      o_angleVec.push_back(sol);

    }
  }

  if(o_angleVec.size() > 0){
    ODEBUG2("computeIK(): "<<o_angleVec.size()<<" solutions found. ");
    return true;
  }
  else{
    ODEBUG1("computeIK(): final step: no solution found. ");
    return false;
  }
}

bool CkineRH1::computeIK(const std::vector<double> &i_goal, std::vector<double> &o_angle)
{
  std::vector<std::vector<double> > angleVec;
  if(computeIK(i_goal, angleVec) == false){
    ODEBUG1("computeIK() with init: final step: no solution found. ");
    return false;
  }

  if(angleVec.size() == 0){
    ODEBUG1("computeIK() with init: no solutions found ");
    return false;
  }
  else
    ODEBUG2("computeIK() with init: "<<angleVec.size()<<" solutions found. ");

  // sort
  sortConfigFromInit(angleVec);

  o_angle = angleVec[0];
  
  return true;
}

bool CkineRH1::computeFK(const std::vector<double> &i_jointAngles)
{
  m_fkFlag = false;

  if(i_jointAngles.size() != m_jointNum){
    ODEBUG1("computeFK(): input joint angle number "<<i_jointAngles.size()
	    <<" != jointNum "<<m_jointNum);
    return false;
  }

  if(m_axisDir.size() != m_jointNum){
    ODEBUG1("computeFK(): input axis number "<<m_axisDir.size()
	    <<" != jointNum "<<m_jointNum);
    return false;
  }

  m_jointAngles = i_jointAngles;

  m_jointPos.clear();
  m_jointOri.clear();

  matrix33 jointOri(tvmet::identity<matrix33>());
  vector3 jointPos(0.0, 0.0, 0.0);
  for(unsigned int i=0; i<m_jointNum; i++){
    vector3 l_vec(0.0, 0.0, m_L[i]);
    vector3 pos(jointPos + jointOri * l_vec);
    jointPos = pos;
    m_jointPos.push_back(jointPos);

    vector3 rotRpy(0.0, 0.0, 0.0);
    rotRpy[m_axisDir[i]] = m_jointAngles[i];
    matrix33 ori(jointOri * rotFromRpy(rotRpy));
    jointOri = ori;
    m_jointOri.push_back(jointOri);
  }

  // add the tcp 
  vector3 pos(jointPos + jointOri * m_tcp);
  m_jointPos.push_back(pos);
  m_jointOri.push_back(jointOri);

  m_fkFlag = true;

  return true;
}

bool CkineRH1::getJointPosFK(unsigned int i_joint, vector3 &pos, matrix33 &rot)
{
  if(i_joint >= m_jointPos.size()){
    ODEBUG1("getJointPosFK(): input angle should be < "<<m_jointNum); 
    return false;
  }
  if(!m_fkFlag){
    ODEBUG1("getJointPosFK(): fk has not been computed."); 
    return false;
  }

  pos = m_jointPos[i_joint];
  rot = m_jointOri[i_joint];

  return true;
}

bool CkineRH1::loadParam(const std::string &i_paramFile)
{
  bool status = true;
  if(!getParamData<double>(i_paramFile, DOF_RH1, LLIMIT_TAG, m_lLimit)){
    ODEBUG1("loadParam(): "<<LLIMIT_TAG<<" not loaded. ");
    status = false;
  }
  if(!getParamData<double>(i_paramFile, DOF_RH1, ULIMIT_TAG, m_uLimit)){
    ODEBUG1("loadParam(): "<<ULIMIT_TAG<<" not loaded. ");
    status = false;
  }

  // set the limitFlag set flag
  m_limitFlag = true;

  if(!getParamData<double>(i_paramFile, DOF_RH1, LINKLENGTH_TAG, m_L)){
    ODEBUG1("loadParam(): "<<LINKLENGTH_TAG<<" not loaded. ");
    status = false;
  }
  std::vector<double> tcp_vec;
  bool flag = getParamData<double>(i_paramFile, 3, TCP_TAG, tcp_vec);
  if(flag)
    m_tcp = vector3(tcp_vec[0], tcp_vec[1], tcp_vec[2]);

  else{
    ODEBUG1("loadParam(): "<<TCP_TAG<<" not loaded. ");
    status = false;
  }

  if(!getParamData<unsigned int>(i_paramFile, DOF_RH1, AXISDIR_TAG, m_axisDir)){
    ODEBUG1("loadParam(): "<<AXISDIR_TAG<<" not loaded. ");
    status = false;
  }

  return status;
}


bool CkineRH1::sortConfigFromInit(std::vector<std::vector<double> > &io_angleVec)
{
  tvmet::Vector<double, DOF_RH1> crntAngle, initCfg;


  for(unsigned int i=0; i<DOF_RH1; i++)
    initCfg[i] = m_initConf[i];

#if DEBUG==2
  std::cerr<<"sortConfigFromInit(): init config: ";
  for(unsigned int i=0; i<DOF_RH1; i++)
    std::cerr<<" "<<initCfg[i];
  std::cerr<<std::endl;
#endif

  std::vector<configPair> configPairVec;
  for(unsigned int i=0; i<io_angleVec.size(); i++){
    for(unsigned int j=0; j<DOF_RH1; j++)
      crntAngle[j] = io_angleVec[i][j];

    configPairVec.push_back(configPair(norm2(crntAngle - initCfg), 
				       io_angleVec[i]));

#if DEBUG==2
    std::cerr<<"sortConfigFromInit(): current: ";
    for(unsigned int j=0; j<DOF_RH1; j++)
      std::cerr<<" "<<crntAngle[j];
    std::cerr<<std::endl;

    std::cerr<<"sortConfigFromInit(): different from current: ";
    for(unsigned int j=0; j<DOF_RH1; j++)
      std::cerr<<" "<<(crntAngle - initCfg)[j];
    std::cerr<<std::endl;
#endif

  }

  std::sort(configPairVec.begin(), configPairVec.end(), compDist());

#if DEBUG==2
  ODEBUG2("sortConfigFromInit(): sort config result: ");
  for(unsigned int i=0; i<configPairVec.size(); i++){
    std::cerr<<"dist "<<configPairVec[i].first<<", config ";
    for(unsigned int j=0; j<DOF_RH1; j++)
      std::cerr<<" "<<configPairVec[i].second[j];
    std::cerr<<std::endl;
  }
#endif

  io_angleVec.clear();
  for(unsigned int i=0; i<configPairVec.size(); i++){
    io_angleVec.push_back(configPairVec[i].second);
  }

  return true;
}

bool CkineRH1::checkLimit(unsigned int i_jnum, double i_val) const
{
  if(!m_limitFlag){
    ODEBUG1("checkLimit(): limit not set ");
    return false;
  }
    
  return ((i_val > m_lLimit[i_jnum]) && (i_val < m_uLimit[i_jnum]));
}

double CkineRH1::limitAngle(double i_angle)
{
  double val = i_angle;
  
  if(i_angle <= PI && i_angle >= -PI)
    return val;

  if(val > PI){
    while(val > PI)
      val -= 2*PI;
  }
  
  if(val < -PI){
    while(val < -PI)
      val += 2*PI;
  }

  return val;
}

void CkineRH1::L(const std::vector<double> &i_L)
{
  assert(m_jointNum == m_L.size());

  m_L = i_L;
}

void CkineRH1::initConf(const std::vector<double> &i_initConf)
{
  assert(m_jointNum == i_initConf.size());

  m_initConf = i_initConf;
}

void CkineRH1::lLimit(const std::vector<double> &i_lLimit)
{
  assert(m_jointNum == i_lLimit.size());

  m_lLimit = i_lLimit;
}

void CkineRH1::uLimit(const std::vector<double> &i_uLimit)
{
  assert(m_jointNum == i_uLimit.size());

  m_uLimit = i_uLimit;
}


template <class T>
bool CkineRH1::getParamData(const std::string &i_fileName, unsigned int i_size,
			  const std::string &i_paramName, std::vector<T> &o_param)
{
  // open file
  std::ifstream fin(i_fileName.c_str());

  if(!fin){
    ODEBUG1("getParamData(): cannot open parameter file "<<i_fileName);
    return false;
  }

  o_param.resize(i_size);

  // find i_paramFile
  char str[1024];

  while(!fin.eof()){

    fin.getline(str, 1024);  // get line
    std::istringstream rsb(str);

    rsb>>std::ws;
    
    std::string tag; 
    rsb>>tag;

    // skip comment
    if(tag.c_str()[0] == '#')
      continue;

    if(tag == i_paramName){
      ODEBUG2("getParamData(): reading parameter "<<tag<<" ... ");
      for(unsigned int i=0; i<i_size; i++){
	rsb >> o_param[i];
      }

#if DEBUG==2
      ODEBUG2("getParamData(): param read:");
      for(unsigned int i=0; i<o_param.size(); i++)
	std::cerr<<" "<<o_param[i];
      std::cerr<<std::endl;
#endif
      return true;
    }
    fin >> std::ws; // skip blank or line feed
  }

  // not found. 
  return false;
}
