// -*- C++ -*-
/*!
 * @file  pathProvider.cpp
 * @brief Module providing path
 * @date $Date$
 *
 * $Id$
 */

#include "pathProvider/pathProvider.h"

#include <fstream>
#include <sstream>
#include <math.h>
#include <assert.h>

#define MIN_TIME (1e-6)

#define BUF_SIZE 50
#define MIN_BUF_SIZE 30

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cerr << "hppRtmPathProvider:" << x << std::endl
#define ODEBUG1(x) std::cerr << "hppRtmPathProvider:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "hppRtmPathProvider:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

double calcDist(TimedDoubleSeq v1, TimedDoubleSeq v2, unsigned int ndof);

// Module specification
// <rtc-template block="module_spec">
static const char* pathprovider_spec[] =
  {
    "implementation_id", "pathProvider",
    "type_name",         "pathProvider",
    "description",       "Module providing path",
    "version",           "1.0.0",
    "vendor",            "AIST",
    "category",          "MotionPlanner",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    "exec_cxt.periodic.rate", "1.0",
    // Configuration variables
    "conf.default.CommandDegFlag", "0",
    "conf.default.CommandOffsetFile", "none",
    "conf.default.PathFile", "none",
    "conf.default.InTimeStep", "0.05",
    "conf.default.OutTimeStep", "0.01",
    "conf.default.MinDist", "0.1",
    "conf.default.OutputMode", "const",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
pathProvider::pathProvider(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_inPathIn("inPath", m_inPath),
    m_inCurrentConfigIn("inCurrentConfig", m_inCurrentConfig),
    m_PathStartIn("PathStart", m_PathStart),
    m_CrntBufSizeIn("CrntBufSize", m_CrntBufSize),
    m_inStopIn("inStop", m_inStop),
    m_outCommandConfigOut("outCommandConfig", m_outCommandConfig),
    m_outCommandConfigSeqOut("outCommandConfigSeq", m_outCommandConfigSeq),
    m_outCurrentConfigOut("outCurrentConfig", m_outCurrentConfig),
    m_outMotionFinishedOut("outMotionFinished", m_outMotionFinished),
    m_outStopOut("outStop", m_outStop),

    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("inPath", m_inPathIn);
  registerInPort("inCurrentConfig", m_inCurrentConfigIn);
  registerInPort("PathStart", m_PathStartIn);
  registerInPort("CrntBufSize", m_CrntBufSizeIn);
  registerInPort("inStop", m_inStopIn);
  
  // Set OutPort buffer
  registerOutPort("outCommandConfig", m_outCommandConfigOut);
  registerOutPort("outCommandConfigSeq", m_outCommandConfigSeqOut);
  registerOutPort("outCurrentConfig", m_outCurrentConfigOut);
  registerOutPort("outMotionFinished", m_outMotionFinishedOut);
  registerOutPort("outStop", m_outStopOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

}

/*!
 * @brief destructor
 */
pathProvider::~pathProvider()
{
}



RTC::ReturnCode_t pathProvider::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("CommandDegFlag", m_CommandDegFlag, "0");
  bindParameter("CommandOffsetFile", m_CommandOffsetFile, "none");
  bindParameter("PathFile", m_PathFile, "none");
  bindParameter("InTimeStep", m_InTimeStep, "0.05");
  bindParameter("OutTimeStep", m_OutTimeStep, "0.01");
  bindParameter("MinDist", m_MinDist, "0.1");
  bindParameter("OutputMode", m_OutputMode, "const");
  
  // </rtc-template>
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t pathProvider::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pathProvider::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pathProvider::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t pathProvider::onActivated(RTC::UniqueId ec_id)
{
  RTC::ReturnCode_t status = RTC::RTC_OK;

  // read Path File to m_inPath
  if(m_PathFile != "none"){
    if(readPath() != RTC::RTC_OK){
      ODEBUG1("reading path file "<<m_PathFile<<" failed. ");
      status = RTC::RTC_ERROR;
    }
    m_crntPathRank = 1;
  }

  // read Offset Data
  if(m_CommandOffsetFile != "none"){
    if(readCommandOffset() != RTC::RTC_OK){
      ODEBUG1("reading command offset file "<<m_CommandOffsetFile<<" failed. ");
      status = RTC::RTC_ERROR;
    }
  }

  ODEBUG2(" hppRtmPathProvider activated. ");
  if(status == RTC::RTC_OK)
    ODEBUG2(" returning OK");

  m_outMotionFinished.data = false;

  return status;
}


/*
RTC::ReturnCode_t pathProvider::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t pathProvider::onExecute(RTC::UniqueId ec_id)
{
  // get path from inport if there is new one
  // read planned path
  // debug
  bool motionFinishedOld = m_outMotionFinished.data;

  if(m_inPathIn.isNew()){
    m_inPathIn.read();

    ODEBUG2(" path read. size "<<m_inPath.data.length());

    setPathTime();
    printPath();

    //  set zero
    // set start
    m_crntPathRank = 1;

    m_outMotionFinished.data = false;
  }

  // get current config
  bool config_new_flag = false;
  if(m_inCurrentConfigIn.isNew()){
    m_inCurrentConfigIn.read();
 
#if DEBUG == 2   
    /*
    std::cerr<<"current config read: ";
    for(unsigned int i=0; i<m_inCurrentConfig.data.length(); i++)
      std::cerr<<" "<<m_inCurrentConfig.data[i];
    std::cerr<<std::endl;
    */
#endif

    // output the current config to outport.
    m_outCurrentConfig = m_inCurrentConfig;
    m_outCurrentConfigOut.write();

    config_new_flag = true;
  }

  // if m_inPath.data.length() == 0; do nothing.
  // the path should include minimally two configs.
  if(m_inPath.data.length() == 0)
    return RTC::RTC_OK;

  // variable mode: next output after the error becomes less than m_MinDist
  if(m_OutputMode == "variable"){
    // see if the config is reached
    // if yes, go to next config
    // after reaching the last config, m_crntPathRank could be equal to m_inPath.data.length()
    unsigned int rank = m_crntPathRank >= m_inPath.data.length() ? m_inPath.data.length()-1 : m_crntPathRank;

    if(config_new_flag && m_inPath.data.length() > 0){
      
      unsigned int ndof = m_inCurrentConfig.data.length() ; 
      double dist = calcDist(m_inPath.data[rank].cfg, m_inCurrentConfig, ndof);
      ODEBUG2(" now dist "<<dist);

      if(dist < m_MinDist){

	if(m_crntPathRank < m_inPath.data.length()){
	  m_crntPathRank++;
	}
      }
      ODEBUG2(" crrentPathRank "<<m_crntPathRank<<", path element num "<<m_inPath.data.length());
#if DEBUG == 2
      std::cout<<": output ";
      for(unsigned int i=0; i<m_inPath.data[rank].cfg.data.length(); i++)
	std::cout<<m_inPath.data[rank].cfg.data[i]/M_PI*180<<" ";
      std::cout<<std::endl;
#endif
    }
    // if no, keep current config

    // outport: current coifig
    if(m_inPath.data.length() > 0){

      if(m_crntPathRank == m_inPath.data.length())
	m_outMotionFinished.data = true;
	
      m_outCommandConfig = m_inPath.data[rank].cfg;
      m_outCommandConfigOut.write();
    }
  }

  // const mode:
  // with output sampleing .
  // input sampling
  else if(m_OutputMode == "const"){
    //     if(config_new_flag){

    bool outputFlag = false;
    // look at the controller buffer size
    if(m_CrntBufSizeIn.isNew()){
      m_CrntBufSizeIn.read();
      ODEBUG2("m_CrntBufSize: "<<m_CrntBufSize.data);
      outputFlag = true;
    }

    if(outputFlag){
    // if the size of the controller is smaller than MIN_BUF_SIZE,
    // then send the block of commands.
      unsigned int bufSize = 0;
      if(m_CrntBufSize.data < MIN_BUF_SIZE){
	m_outCommandConfigSeq.data.length(BUF_SIZE);
	for(unsigned int i=0; i<BUF_SIZE; i++){
	  if(m_OutConfigQueue.size() == 0)
	    fillConfigQueue();

	  // get the sequence for output
	  if(m_OutConfigQueue.size() == 0){ // still zero: empty
	    ODEBUG2(" queue still zero, break ");
	    
	    // set the flag finished.
	    if(m_CrntBufSize.data == 0)
	      m_outMotionFinished.data = true;
	    break;
	  }

	  m_outCommandConfigSeq.data[i].cfg = m_OutConfigQueue[0];
	  m_outCommandConfigSeq.data[i].time = i*m_OutTimeStep;
	  m_OutConfigQueue.pop_front();
	  bufSize++;

#if DEBUG == 2
	  std::cout<<"output config "<<i<<": ";
	  for(unsigned int j=0; j<m_outCommandConfigSeq.data[i].cfg.data.length(); j++)
	    std::cout<<" "<<m_outCommandConfigSeq.data[i].cfg.data[j];
	  std::cout<<std::endl;
#endif
	}

	if(bufSize > 0){
	  // resize the buffer
	  ODEBUG2("buf size "<<bufSize);
	  m_outCommandConfigSeq.data.length(bufSize);
	  m_outCommandConfigSeqOut.write();
// debug
#if DEBUG == 2
	  for(unsigned int i=0; i<m_outCommandConfigSeq.data.length(); i++){
	    CfgElement ce = m_outCommandConfigSeq.data[i];
	    m_foutSeq<<ce.time;
	    for(unsigned int j=0; j<ce.cfg.data.length(); j++){
	      m_foutSeq<<"\t"<<ce.cfg.data[j];
	    }
	    m_foutSeq<<std::endl;
	  }
#endif
	}

	ODEBUG2(" ============== ");
	// } // config_new_flag
      }
    }
  }

  // stop
  // clear the queue, set the current config as a reference
  // set finished flag
  if(m_inStopIn.isNew()){
    m_inStopIn.read();

    
    if(m_inStop.data == true){

      // clear the path input
      m_inPath.data.length(0);

      ODEBUG1(" stop signal received. ");

      // output stop signal
      m_outStop.data = true;
      m_outStopOut.write();

      if(m_OutputMode == "variable"){
	// direct mode

	// get and send the current state
	m_outCommandConfig = m_inCurrentConfig;
	m_outCommandConfigOut.write();
      }
    
      else if(m_OutputMode == "const"){
	// buffer mode
	// wait for the 
	m_outCommandConfigSeq.data.length(1);
	m_outCommandConfigSeq.data[0].cfg = m_inCurrentConfig;
	m_outCommandConfigSeq.data[0].time = 0;

	// send the current state
	m_outCommandConfigSeqOut.write();

	// clear the queue
	m_OutConfigQueue.clear();
      }

      m_outMotionFinished.data = true;
      
      // reinitialize variables
      m_crntPathRank = 1;
    }

  }

  if(motionFinishedOld == false && m_outMotionFinished.data == true){
    ODEBUG1(" ============= outMotionFinished set to  "<<m_outMotionFinished.data
	     <<" ============= ");
  }
  m_outMotionFinishedOut.write();

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t pathProvider::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pathProvider::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pathProvider::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pathProvider::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t pathProvider::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t pathProvider::readPath()
{
  if(m_PathFile == "none" || m_PathFile == ""){
    ODEBUG2(" no path file specified ");
    return RTC::RTC_OK;
  }

  std::ifstream fin(m_PathFile.c_str());

  if(!fin){
    ODEBUG1("cannot open "<<m_PathFile);
    return RTC::RTC_ERROR;
  }

  std::vector<std::vector<double> > pathVec;

  while(!fin.eof()){

    // read from file
    char str[1024];

    fin.getline(str, 1024);  // get line
    std::istringstream rsb(str);

    // get vector
    std::vector<double> vec;
    vec.clear();
    double v;
    while(rsb>>v){
      vec.push_back(v);
    }

    pathVec.push_back(vec);
    // make path
    fin >> std::ws; // skip blank or line feed
  }

  m_inPath.data.length(pathVec.size());
  for(unsigned int i=0; i<pathVec.size(); i++){
    m_inPath.data[i].cfg.data.length(pathVec[i].size());

    for(unsigned int j=0; j<pathVec[i].size(); j++)
      m_inPath.data[i].cfg.data[j] = pathVec[i][j];
  }
  setPathTime();

  printPath();

  return RTC::RTC_OK;
}

RTC::ReturnCode_t pathProvider::readCommandOffset()
{
  if(m_CommandOffsetFile == "none" || m_CommandOffsetFile == ""){
    ODEBUG2(" no path file specified ");
    return RTC::RTC_OK;
  }

  std::ifstream fin(m_CommandOffsetFile.c_str());

  if(!fin){
    ODEBUG1("cannot open "<<m_CommandOffsetFile);
    return RTC::RTC_ERROR;
  }

  m_CommandOffsetVec.clear();

  while(!fin.eof()){

    // get vector
    double v;

    fin>>v;
    m_CommandOffsetVec.push_back(v);

    fin >> std::ws; // skip blank or line feed
  };

#if DEBUG == 2
  std::cout<<"Command offset read:";
  for(unsigned int i=0; i<m_CommandOffsetVec.size(); i++)
    std::cout<<" "<<m_CommandOffsetVec[i];
  std::cout<<std::endl;
#endif

  return RTC::RTC_OK;
}

RTC::ReturnCode_t pathProvider::fillConfigQueue()
{
  ODEBUG2("New queue at "<<m_crntPathRank<<", inPath length "
	   <<m_inPath.data.length());

  // if no queue left, go to next
  if(m_crntPathRank < m_inPath.data.length()-1){

    // initialize at zero
    RTC::TimedDoubleSeq cfgSeq_start, cfgSeq_end;
    double t_start, t_end;
    if(m_crntPathRank > 0){
      cfgSeq_start = m_inPath.data[m_crntPathRank-1].cfg;
      t_start = m_inPath.data[m_crntPathRank-1].time;
    }
    else{
      ODEBUG1("Error: pathProvider::onExecute(): m_crntPathRank "
	       <<m_crntPathRank<<" should be greater than 0 ");
      return RTC::RTC_ERROR;
    }

    cfgSeq_end = m_inPath.data[m_crntPathRank].cfg;
    t_end = m_inPath.data[m_crntPathRank].time;
    unsigned int cfgSize = cfgSeq_end.data.length();

    // get the samplings with m_OutTimeStep
    unsigned int num_samples 
      = (unsigned int) ((t_end-t_start+MIN_TIME)/m_OutTimeStep);

    m_OutConfigQueue.clear();
    RTC::TimedDoubleSeq cfg_step;
    cfg_step.data.length(cfgSize);
    for(unsigned int j=0; j<cfgSize; j++)
      cfg_step.data[j] = (cfgSeq_end.data[j]-cfgSeq_start.data[j])/num_samples;
	 
    for(unsigned int i=1; i<=num_samples; i++){ // start with 1
      RTC::TimedDoubleSeq cfg;
      cfg.data.length(cfgSize);
      for(unsigned int j=0; j<cfg.data.length(); j++)
	cfg.data[j] = cfgSeq_start.data[j]+cfg_step.data[j]*i;
      m_OutConfigQueue.push_back(cfg);
    }

    m_crntPathRank++;
  }
  // see if no queue left
  ODEBUG2("config queue size "<<m_OutConfigQueue.size()
	   <<" at m_crntPathRank "<<m_crntPathRank);

  return RTC::RTC_OK;
}

void pathProvider::setPathTime()
{
  for(unsigned int i=0; i<m_inPath.data.length(); i++)
    m_inPath.data[i].time = i*m_InTimeStep;
}

void pathProvider::printPath()
{
  std::cout<<"Path: "<<std::endl;
  for(unsigned int i=0; i<m_inPath.data.length(); i++){
    std::cout<<" t= "<<m_inPath.data[i].time<<":";
    for(unsigned int j=0; j<m_inPath.data[i].cfg.data.length(); j++)
      std::cout<<" "<<m_inPath.data[i].cfg.data[j];
    std::cout<<std::endl;
  }
}

double calcDist(TimedDoubleSeq v1, TimedDoubleSeq v2, unsigned int ndof)
{
  assert(v1.data.length() == v2.data.length());

  double dist = 0;

  unsigned int dof = v1.data.length() < ndof ? v1.data.length() : ndof;
  ODEBUG2("diff: ");
  for(unsigned int i=0; i<dof; i++){
    dist += (v1.data[i]-v2.data[i])*(v1.data[i]-v2.data[i]);
    ODEBUG2(v1.data[i]-v2.data[i]<<" ");
  }
  ODEBUG2("");

  return sqrt(dist);
}

extern "C"
{
 
  void pathProviderInit(RTC::Manager* manager)
  {
    RTC::Properties profile(pathprovider_spec);
    manager->registerFactory(profile,
                             RTC::Create<pathProvider>,
                             RTC::Delete<pathProvider>);
  }
  
};


