// -*- C++ -*-
/*!
 * @file  trajGen.cpp
 * @brief RTC for path timing planner
 * @date $Date$
 *
 * $Id$
 */

#include "trajGen/trajGen.h"
#include "trajGen/trajGenTools.h"

// Module specification
// <rtc-template block="module_spec">
static const char* trajgen_spec[] =
  {
    "implementation_id", "trajGen",
    "type_name",         "trajGen",
    "description",       "RTC for path timing planner",
    "version",           "1.0.0",
    "vendor",            "AIST",
    "category",          "MotionPlanner",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    "exec_cxt.periodic.rate", "1.0",
    // Configuration variables
    "conf.default.samplingTime", "0.044",
    "conf.default.saveDataName", "none",
    "conf.default.maxJointVel", "0.785398,0.523304,0.409062,0.785398,0.785398,0.785398,0.013317",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
trajGen::trajGen(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_inTrajectoryIn("inTrajectory", m_inTrajectory),
    m_outTrajectoryOut("outTrajectory", m_outTrajectory),

    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("inTrajectory", m_inTrajectoryIn);
  
  // Set OutPort buffer
  registerOutPort("outTrajectory", m_outTrajectoryOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

}

/*!
 * @brief destructor
 */
trajGen::~trajGen()
{
}



RTC::ReturnCode_t trajGen::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("samplingTime", m_samplingTime, "0.044");
  bindParameter("saveDataName", m_saveDataName, "none");
  bindParameter("maxJointVel", m_maxJointVel, "0.785398,0.523304,0.409062,0.785398,0.785398,0.785398,0.013317");
  
  // </rtc-template>
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t trajGen::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t trajGen::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t trajGen::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t trajGen::onActivated(RTC::UniqueId ec_id)
{
  // set the velocity limit
  m_uvlimit  = getVectorDelim(m_maxJointVel, ",");

  m_lvlimit.resize(m_uvlimit.size());
  for(unsigned int i=0; i<m_uvlimit.size(); i++)
    m_lvlimit[i] = -m_uvlimit[i];

  std::cerr<<" joint limit set: (size "<<m_lvlimit.size()<<")"<<std::endl;
  for(unsigned int i=0; i<m_uvlimit.size(); i++){
    std::cerr<<" joint "<<i<<": ["<<m_lvlimit[i]<<", "<<m_uvlimit[i]<<"]"
	     <<std::endl;
  }

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t trajGen::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t trajGen::onExecute(RTC::UniqueId ec_id)
{
  // get the value
  if(m_inTrajectoryIn.isNew()){
    m_inTrajectoryIn.read();

    std::cout<<"trajectory received. size "<<m_inTrajectory.data.length()<<std::endl;

    std::vector<std::vector<double> > mjTrajVec;
    if(!calcMinJerkTrajectory(m_inTrajectory, m_samplingTime, m_uvlimit, m_lvlimit,
			      mjTrajVec)){
      return RTC::RTC_ERROR;
    }

    vec2TrajData(mjTrajVec, m_outTrajectory, m_samplingTime);

    std::cout<<" trajectory succeeded: result: "<<std::endl;
    printPath(std::cout, m_outTrajectory);

    // save the data
    if(m_saveDataName != "none"){
      char fname_time[256], fname[512];
      getStrDateTime(fname_time);
      sprintf(fname, "%s", m_saveDataName.c_str());
      std::ofstream fout(fname);
      if(fout){
	printPath(fout, m_outTrajectory);
      }
      else{
	std::cerr<<" cannot open "<<fname<<std::endl;
      }
    }

    // output
    m_outTrajectoryOut.write();

    // get sampling time
    // if sparse, make minjerk, then optimize
    // if sampled, optimize

  }

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t trajGen::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t trajGen::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t trajGen::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t trajGen::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t trajGen::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void trajGenInit(RTC::Manager* manager)
  {
    RTC::Properties profile(trajgen_spec);
    manager->registerFactory(profile,
                             RTC::Create<trajGen>,
                             RTC::Delete<trajGen>);
  }
  
};


