
#include "trajGen/trajGenTools.h"
#include "trajGen/trajGen.h"

#include <math.h>

#define TIME_RATIO 1.2  // safety ratio for time

// compute minimum jerk trajectory
// suppose a sparse path input.
// get velocity limit, otherwise a fixed value
// Max velocity of minjerk curve is:

bool calcMinJerkTrajectory(const TrajData &i_traj, double i_samplingTime, 
			   const std::vector<double> &i_uvlimit, 
			   const std::vector<double> &i_lvlimit, 
			   std::vector<std::vector<double> > &o_trajVec)
{
  unsigned int ndof = 0;
  for(unsigned int i=0; i<i_traj.data.length()-1; i++){
    
    // get the maximal time from the displacement.
    if(ndof == 0)
      ndof = i_traj.data[i].cfg.data.length();
    else{
      if(ndof != i_traj.data[i].cfg.data.length()){
	std::cerr<<"calcMinJerkTrajectory(): nof number not match at "<<i<<" th path elem. "<<std::endl;
	return false;
      }
    }
    double time_traj;
    if(!calcExecTime(i_traj.data[i].cfg, i_traj.data[i+1].cfg, i_samplingTime, i_uvlimit, i_lvlimit, time_traj))
      return false;

    std::cout<<"time_traj for path elem "<<i<<": "<<time_traj<<std::endl;
    if(!calcMinJerkElemTraj(i_traj.data[i].cfg, i_traj.data[i+1].cfg, time_traj, i_samplingTime, o_trajVec))
      return false;
    
  }
  // now o_trajVec includes all the joint angles.
  std::cout<<" o_trajVec size "<<o_trajVec.size()<<std::endl;
  return true;
   
}


bool calcMinJerkTrajectory(const TrajData &i_traj, double i_samplingTime, 
			   const std::vector<double> &i_uvlimit, 
			   const std::vector<double> &i_lvlimit, 
			   TrajData &o_traj)
{
  std::vector<std::vector<double> > o_trajVec;
  if(!calcMinJerkTrajectory(i_traj, i_samplingTime, i_uvlimit, i_lvlimit, o_trajVec))
     return false;

  vec2TrajData(o_trajVec, o_traj, i_samplingTime);

  std::cout<<" min jerk path generated: "<<std::endl;
  printPath(std::cout, o_traj);

  return true;
}


bool calcExecTime(const RTC::TimedDoubleSeq &i_cfg_init, const RTC::TimedDoubleSeq &i_cfg_final,
		  double i_samplingTime,
		  const std::vector<double> &i_uvlimit, const std::vector<double> &i_lvlimit,
		  double &o_time)
		    
{
  // for Minimum jerk trajectory, the maximum velocity is
  // (15d)/(8D), at t = T/2 when the initial/final velocities and accelerations are zero.
  // where d = the traveling distance and D is tf - t0.
  // therefore, the traveling time is maximum value (15d)/(8vmax) for all joints.
  double max_time = -1;
  if(i_cfg_init.data.length() != i_cfg_final.data.length()){
    std::cerr<<"calcExecTime() : dof of init "<<i_cfg_init.data.length()<<" and final "
	     <<i_cfg_final.data.length()<<" not the same "<<std::endl;
    return false;
  }
  for(unsigned int i=0; i<i_cfg_init.data.length(); i++){
    double t;
    double diff = i_cfg_final.data[i] - i_cfg_init.data[i];
    double lim = diff > 0 ? i_uvlimit[i] : i_lvlimit[i];

    t = (fabs(diff)*15)/(fabs(lim)*8)*TIME_RATIO;
    max_time = (max_time > t) ? max_time : t;
  }

  unsigned long num = (unsigned long) (ceil(max_time/i_samplingTime));
  
  o_time = num * i_samplingTime;

  return true;
}

bool calcMinJerkElemTraj(const RTC::TimedDoubleSeq &i_cfg_init, const RTC::TimedDoubleSeq &i_cfg_final,
			 double i_time_traj, double i_samplingTime, 
			 std::vector<std::vector<double> > &o_trajVec)
{
  // get the time with the sampling time
  if(i_cfg_init.data.length() != i_cfg_final.data.length()){
    std::cerr<<"calcMinJerkElemTraj() : dof of init "<<i_cfg_init.data.length()<<" and final "
	     <<i_cfg_final.data.length()<<" config not the same"<<std::endl;
    return false;
  }
  
  std::vector<std::vector<double> > indvTraj;
  unsigned int traj_size = 0;
  unsigned int cfg_size = i_cfg_init.data.length();

  for(unsigned int i=0; i<cfg_size; i++){
    std::vector<double> traj;
    minJerkInterpolation(i_time_traj, i_cfg_init.data[i], i_cfg_final.data[i], i_samplingTime, traj); 
    indvTraj.push_back(traj);

    if(traj_size == 0)
      traj_size = traj.size();
    else{
      if(traj_size != traj.size()){
	std::cerr<<"calcMinJerkElemTraj() : trajectory size error "<<std::endl;
      }
    }
  }

  for(unsigned int i=0; i<traj_size; i++){
    std::vector<double> config(cfg_size);
    for(unsigned int j=0; j<cfg_size; j++){
      config[j] = indvTraj[j][i];
    }
    o_trajVec.push_back(config);
  }
  return true;
}

void minJerkInterpolation(double tot_time, double x_init, double x_fin, double samplingTime, std::vector<double> &seq_x)
{
  double coeff[6];
  double tau, val, val_prev = 0, vel;
  seq_x.clear();
  //seq_dx.clear();

  double t_init = 0.0, t_fin = tot_time;
  double v_init = 0.0;    
  double acc_init = 0.0;  
  
  //
  double D = t_fin - t_init;
  double dist = x_fin - x_init;

  coeff[0] = x_init;   
  coeff[1] = D*v_init;   
  coeff[2] = D*D*acc_init/2;   
  
  coeff[3] = (-3*D*D)/2 * acc_init - 6*D*v_init + 10*dist;   
  coeff[4] =  (3*D*D)/2 * acc_init + 8*D*v_init - 15*dist;   
  coeff[5] =   -(D*D)/2 * acc_init - 3*D*v_init +  6*dist;   
  //

  for(int i=0; samplingTime*i < tot_time+MIN_DIST; i++){
    tau = i*samplingTime / tot_time;
    val = coeff[0];

    for(int j=1; j<6; j++){
      val += coeff[j] * pow(tau, (double)(j));
    }
    
    if(i == 0)
      vel = 0;
    else
      vel = (val - val_prev) / samplingTime;

    val_prev = val;

    seq_x.push_back(val);
    //seq_dx.push_back(vel);
  }
}

void printPath(std::ostream& outs, const TrajData &path)
{
  // out<<"Path: "<<std::endl;
  for(unsigned int i=0; i<path.data.length(); i++){
    // std::cout<<" t= "<<
    // outs<<path.data[i].time<<" ";
    for(unsigned int j=0; j<path.data[i].cfg.data.length(); j++)
      outs<<path.data[i].cfg.data[j]<<" ";
    outs<<std::endl;
  }
}

void vec2TrajData(const std::vector<std::vector<double> > &i_vec, TrajData &o_traj, 
		  double i_samplingTime)
{
  o_traj.data.length(i_vec.size());
  for(unsigned int i=0; i<i_vec.size(); i++){
    unsigned int ndof = i_vec[i].size();
    o_traj.data[i].cfg.data.length(ndof);
    o_traj.data[i].time = i*i_samplingTime;
    for(unsigned int j=0; j<ndof; j++){
      o_traj.data[i].cfg.data[j] = i_vec[i][j];
    }
  }
}

void TrajData2vec(const TrajData &i_traj, std::vector<std::vector<double> > &o_vec)
{
  o_vec.clear();

  for(unsigned int i=0; i<i_traj.data.length(); i++){
    unsigned int ndof = i_traj.data[i].cfg.data.length();
    std::vector<double> vec(ndof); 
    
    for(unsigned int j=0; j<ndof; j++){
      vec[j] = i_traj.data[i].cfg.data[j];
    }
    o_vec.push_back(vec);
  }
}

void getStrDateTime(char* fname_time)
{
  struct tm *date;
  time_t now;
  int year, month, day;
  int hour, minute, second;

  // get time
  time(&now);
  date = localtime(&now);

  year = date->tm_year + 1900;
  month = date->tm_mon + 1;
  day = date->tm_mday;

  hour = date->tm_hour;
  minute = date->tm_min;
  second = date->tm_sec;

  sprintf(fname_time, "%04d%02d%02d-%02d%02d%02d", 
	  year, month, day, hour, minute, second);

}

std::vector<double> getVectorDelim(const std::string &i_line, const std::string &i_delim)
{
  std::vector<double> vec;
  std::string::size_type begin = 0;
  std::string::size_type end = 0;
  while ( (begin = i_line.find_first_not_of(i_delim, end)) != std::string::npos ) {
    end = i_line.find_first_of(i_delim, begin);
    vec.push_back(atof(i_line.substr(begin, end-begin).c_str()));
  }
  return vec;
}


